# This file contains definitions for dismemr-related macros
# and functions that we use during configure stage.

# Check for GasNet
MACRO (CHECK_DISTMEM_GASNET)
  SET (ENABLE_DISTMEM OFF)

  IF (NOT ${CMAKE_SYSTEM_NAME} STREQUAL "Linux")
    SET (DISTMEM OFF)
    MESSAGE(STATUS "Distmem/Shray only supports Linux")
  ENDIF ()

  IF (DISTMEM)
    SET(ENABLE_DISTMEM ON)

    IF (NOT (MPI OR IBV OR UDP))
        MESSAGE(FATAL_ERROR "Please set a gasnet conduit (mpi, ibv, udp) to ON.")
    ENDIF ()

    IF (MPI)
        FIND_PACKAGE(MPI)
        IF (${MPI_FOUND})
            SET(GASNet_CONDUIT mpi)
        ELSE ()
            MESSAGE(STATUS "No MPI installation, distmem backend is not built")
            SET(ENABLE_DISTMEM OFF)
            SET(DISTMEM OFF)
            SET(MPI OFF)
        ENDIF ()
    ENDIF ()

    IF (IBV)
        SET(GASNet_CONDUIT ibv)
    ENDIF ()

    IF (UDP)
        SET(GASNet_CONDUIT udp)
    ENDIF ()
  ENDIF ()

  IF (DISTMEM)
    SET(GASNet_THREADING par)

    # Using cmake's automagic submodule system is brittle, so we add Shray as an
    # external project, copy over the header, and set the path in 
    # setup/sac2crc.backend.distmem
    MESSAGE(STATUS "Adding Shray as external project")
    SET(SHRAY_SOURCE_DIR "${PROJECT_SOURCE_DIR}/ext/shray2")
    ExternalProject_Add(shray_library
        DOWNLOAD_COMMAND "" # this is to prevent any download target from being called
        INSTALL_COMMAND "" # this is to prevent any install target from being called
        PREFIX runtime_build
        SOURCE_DIR ${SHRAY_SOURCE_DIR}
        BUILD_ALWAYS 1 # this is so we react on changes within the source tree
        BINARY_DIR ${SHRAY_BINARY_DIR}
        CMAKE_ARGS
            -DMPIEXEC_EXECUTABLE=${MPIEXEC_EXECUTABLE}
            -DMPI_C_COMPILER=${MPI_C_COMPILER}
            -DGASNet_ROOT_DIR=${GASNet_ROOT_DIR}
            -DGASNet_CONDUIT=${GASNet_CONDUIT}
            -DGASNet_THREADING=${GASNet_THREADING}
            -DCMAKE_BUILD_TYPE=${CMAKE_BUILD_TYPE}
    )

    MESSAGE(STATUS "Copy over Shray header")
    FILE(COPY "${PROJECT_SOURCE_DIR}/ext/shray2/include/shray2/shray.h"
                DESTINATION "${PROJECT_BINARY_DIR}/include")

    MESSAGE(STATUS "Adding target distmem")
    LIST (APPEND RT_TARGETS distmem)

  ENDIF ()

ENDMACRO ()
