/*****************************************************************************
 *
 * file:   identify_side_effect_function_aps.c
 *
 * prefix: DMISEFA
 *
 * IMPORTANT:
 *  This needs to be run after dmisef
 *
 * description:
 *
 *  Marks function applications that have side effects 
 *  (AP_DISTMEMHASSIDEEFFECTS) so codegen can make sure this is only executed
 *  by node 0. This has been marked in the fundef in DMISEF.
 *
 *  We use the term side effect for functions that affect the observable world,
 *  and must therefore only be executed once.
 *
 *  Transforms
 *    x = f(y);
 *  to
 *    x = f(y); // side-effects
 *
 * implementation:
 *  Trivial.
 *
 *****************************************************************************/

#define DBUG_PREFIX "DMISEFA"
#include "debug.h"

#include "identify_side_effect_function_aps.h"

#include "traverse.h"   /* TRAV- */
#include "tree_basic.h" /* Son and attribute access macro's */

node *
DMISEFAap (node *arg_node, info *arg_info)
{
    DBUG_ENTER ();

    AP_DISTMEMHASSIDEEFFECTS (arg_node) =
            FUNDEF_DISTMEMHASSIDEEFFECTS (AP_FUNDEF (arg_node));

    DBUG_PRINT ("Function application %s %s side effects",
                FUNDEF_NAME (AP_FUNDEF (arg_node)),
                (AP_DISTMEMHASSIDEEFFECTS (arg_node)) ? "has" : "has no");

    DBUG_RETURN (arg_node);
}

node *
DMISEFAdoIdentifySideEffectFunctionApplications (node *syntax_tree)
{
    DBUG_ENTER ();

    DBUG_ASSERT (NODE_TYPE (syntax_tree) == N_module,
                "called with non-module node");

    TRAVpush (TR_dmisefa);

    syntax_tree = TRAVdo (syntax_tree, NULL);

    TRAVpop ();

    DBUG_RETURN (syntax_tree);
}

#undef DBUG_PREFIX
