#include "serialize_info.h"

#define DBUG_PREFIX "UNDEFINED"
#include "debug.h"

#include "symboltable.h"
#include "filemgr.h"
#include "tree_basic.h"
#include "namespaces.h"
#include "serialize_symboltable.h"
#include "globals.h"

static void
GenerateSerSymbolTableAdd (stsymbol_t *symbol, stentry_t *entry, FILE *file)
{
    DBUG_ENTER ();

    fprintf (file, "STadd( \"%s\", %d, \"%s\", %d, result, %d);\n", STsymbolName (symbol),
             STsymbolVisibility (symbol), STentryName (entry), STentryType (entry),
             STentryArgc (entry));

    DBUG_RETURN ();
}

static void
GenerateSerSymbolTableHead (node *module, FILE *file)
{
    DBUG_ENTER ();

    fprintf (file, "/* generated by sac2c %s */\n\n", global.version_id);

    fprintf (file, "#include \"sac_serialize.h\"\n\n");

    fprintf (file, "#ifdef __cplusplus\n"
                   "extern \"C\"\n"
                   "#endif\n");

    fprintf (file, "void *__%s__SYMTAB( void)\n", NSgetName (MODULE_NAMESPACE (module)));

    fprintf (file, "{\nvoid *result;\n");
    fprintf (file, "result = STinit();\n");

    DBUG_RETURN ();
}

static void
GenerateSerSymbolTableTail (node *module, FILE *file)
{
    DBUG_ENTER ();

    fprintf (file, "return( result);\n");
    fprintf (file, "}\n");

    DBUG_RETURN ();
}

static void
SerializeSymbolTableSymbol (stsymbol_t *symbol, sttable_t *table, FILE *file)
{
    stentryiterator_t *iterator;

    DBUG_ENTER ();

    iterator = STentryIteratorGet (STsymbolName (symbol), table);

    while (STentryIteratorHasMore (iterator)) {
        GenerateSerSymbolTableAdd (symbol, STentryIteratorNext (iterator), file);
    }

    iterator = STentryIteratorRelease (iterator);

    DBUG_RETURN ();
}

void
SSTserializeSymbolTable (node *module, sttable_t *table)
{
    stsymboliterator_t *iterator;
    FILE *file;

    DBUG_ENTER ();

    file = FMGRwriteOpen ("%s/symboltable.c", global.tmp_dirname);

    GenerateSerSymbolTableHead (module, file);

    iterator = STsymbolIteratorGet (table);

    while (STsymbolIteratorHasMore (iterator)) {
        SerializeSymbolTableSymbol (STsymbolIteratorNext (iterator), table, file);
    }

    iterator = STsymbolIteratorRelease (iterator);

    GenerateSerSymbolTableTail (module, file);

    fclose (file);

    DBUG_RETURN ();
}

#undef DBUG_PREFIX
