/*   Most of this is just a wrapper around h-macro's, but for updating
 *   the mirror variables in the AKD case, it is necessary to generate
 *   assignments for each dimension, so we do that in icm2c_wl.c. */

#define SAC_DIST_REDUCE__SCL(var_NT, foldfun)                                 \
{                                                                             \
    SAC_TR_DISTMEM_PRINT ("Distributed scalar fold on %s", NT_STR (var_NT));  \
    SAC_NT_CBASETYPE(var_NT) *reductionBuffer;                                \
    SAC_HM_MALLOC(reductionBuffer, ShraySize() *                              \
            sizeof(SAC_NT_CBASETYPE(var_NT)), SAC_NT_CBASETYPE(var_NT));      \
    SAC_DISTMEM_Allgather(reductionBuffer, SAC_ADDRESS(var_NT),               \
                sizeof(SAC_NT_CBASETYPE(var_NT)));                            \
    for (size_t i = 0; i < ShraySize(); i++) {                                \
        if (i == ShrayRank()) continue;                                       \
        foldfun(SAC_ADDRESS(var_NT), reductionBuffer[i],                      \
                SAC_ND_A_FIELD(var_NT));                                      \
    }                                                                         \
    SAC_HM_FREE(reductionBuffer);                                             \
}

/* TODO The two Allgathers should be made nonblocking so we only have to
 * synchronise once.
 * We increase the reference count because we manage the memory used
 * in the descriptors, and we do not want the fold function to free them. */
#define SAC_DIST_REDUCE__AKS(var_NT, foldfun)                                 \
{                                                                             \
    SAC_TR_DISTMEM_PRINT("Distributed AKS fold on %s", NT_STR(var_NT));       \
    SAC_NT_CBASETYPE(var_NT) *reductionBuffer;                                \
    SAC_HM_MALLOC(reductionBuffer, ShraySize() *                              \
            SAC_ND_A_SIZE(var_NT) * sizeof(SAC_NT_CBASETYPE(var_NT)),         \
            SAC_NT_CBASETYPE(var_NT));                                        \
    SAC_ND_A_RC__C99(var_NT) += 1;                                            \
    SAC_array_descriptor_t *descs;                                            \
    SAC_HM_MALLOC(descs, BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT)) *            \
                        ShraySize(), SAC_array_descriptor_t);                 \
    SAC_DISTMEM_Allgather(reductionBuffer, SAC_ADDRESS(var_NT),               \
                SAC_ND_A_SIZE(var_NT) *                                       \
                    sizeof(SAC_NT_CBASETYPE(var_NT)));                        \
    SAC_DISTMEM_Allgather(descs, SAC_ND_A_DESC(var_NT),                       \
                          BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT)));           \
    SAC_ND_A_RC__C99(var_NT) -= 1;                                            \
    for (size_t i = 0; i < ShraySize(); i++) {                                \
        if (i == ShrayRank()) continue;                                       \
        foldfun(&SAC_ADDRESS(var_NT), &SAC_ND_A_DESC(var_NT),                 \
                reductionBuffer + i * SAC_ND_A_SIZE(var_NT),                  \
                (SAC_array_descriptor_t)((uintptr_t)descs +                   \
                        i * BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT))),         \
                SAC_ND_A_FIELD(var_NT), SAC_ND_A_DESC(var_NT));               \
    }                                                                         \
    SAC_HM_FREE(reductionBuffer);                                             \
    SAC_HM_FREE(descs);                                                       \
}

/* We do not update the mirror variables */
#define SAC_DIST_REDUCE__AKD_AUD(var_NT, foldfun)                             \
{                                                                             \
    SAC_TR_DISTMEM_PRINT("Distributed AKD/AUD fold on %s", NT_STR(var_NT));   \
    SAC_NT_CBASETYPE(var_NT) *reductionBuffer;                                \
    SAC_ND_A_RC__C99(var_NT) += 1;                                            \
    SAC_array_descriptor_t *descs;                                            \
    SAC_HM_MALLOC(descs, BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT)) *            \
                        ShraySize(), SAC_array_descriptor_t);                 \
    SAC_DISTMEM_Allgather(descs, SAC_ND_A_DESC(var_NT),                       \
                          BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT)));           \
    size_t my_sz = SAC_ND_A_SIZE(var_NT) * sizeof(SAC_NT_CBASETYPE(var_NT));  \
    size_t sizes[ShraySize()];                                                \
    size_t total_size = 0;                                                    \
    SAC_DISTMEM_Allgather(sizes, &my_sz, sizeof(intptr_t));                   \
    for (unsigned int i = 0; i < ShraySize(); i++) {                          \
        total_size += sizes[i];                                               \
        SAC_TR_DISTMEM_PRINT("Size of array %d is %zu bytes", i, sizes[i]);   \
    }                                                                         \
    SAC_HM_MALLOC(reductionBuffer, total_size, SAC_NT_CBASETYPE (var_NT));    \
    SAC_DISTMEM_Allgatherv(ShraySize(), reductionBuffer,                      \
                           SAC_ADDRESS(var_NT), sizes);                       \
    SAC_ND_A_RC__C99(var_NT) -= 1;                                            \
    void *data_ptr = reductionBuffer;                                         \
    void *desc_ptr = descs;                                                   \
    for (size_t i = 0; i < ShraySize(); i++) {                                \
        if (i != ShrayRank()) {                                               \
            foldfun(&SAC_ADDRESS(var_NT), &SAC_ND_A_DESC(var_NT),             \
                    SAC_ND_A_FIELD(var_NT), SAC_ND_A_DESC(var_NT),            \
                    data_ptr, (SAC_array_descriptor_t)desc_ptr);              \
        }                                                                     \
        data_ptr = (void *)((uintptr_t)data_ptr + sizes[i]);                  \
        desc_ptr = (void *)((uintptr_t)desc_ptr +                             \
                            BYTE_SIZE_OF_DESC(SAC_ND_A_DIM(var_NT)));         \
    }                                                                         \
    SAC_HM_FREE(reductionBuffer);                                             \
    SAC_HM_FREE(descs);                                                       \
}
